import { app } from "./app";
import { APIException, NetworkConnectionException, ReactLifecycleException, RuntimeException } from "./Exception";
export class EventLogger {
    constructor() {
        this.environmentContext = {};
        this.uuidCounter = new Date().getTime();
        this.logQueue = [];
    }
    setContext(context) {
        this.environmentContext = context;
    }
    log(_, info = {}) {
        if (typeof _ === "string") {
            return this.appendLog(_, "OK", info);
        }
        else if (_ instanceof NetworkConnectionException) {
            return this.appendLog("networkFailure", "WARN", { errorMessage: _.message });
        }
        else {
            const exceptionInfo = { errorMessage: _.message };
            let errorCode = "error";
            if (_ instanceof APIException) {
                errorCode = `apiError:${_.statusCode}`;
                exceptionInfo.requestURL = _.requestURL;
                exceptionInfo.statusCode = _.statusCode.toString();
            }
            else if (_ instanceof ReactLifecycleException) {
                errorCode = "lifecycleError";
                exceptionInfo.stackTrace = _.componentStack;
                exceptionInfo.appState = JSON.stringify(app.store.getState().app);
            }
            else if (_ instanceof RuntimeException) {
                errorCode = "jsError";
                exceptionInfo.stackTrace = JSON.stringify(_.errorObject);
                exceptionInfo.appState = JSON.stringify(app.store.getState().app);
            }
            return this.appendLog(errorCode, "ERROR", exceptionInfo);
        }
    }
    collect() {
        return this.logQueue;
    }
    empty() {
        this.logQueue = [];
    }
    appendLog(actionOrErrorCode, result, info) {
        const completeContext = {};
        Object.entries(this.environmentContext).map(([key, value]) => {
            completeContext[key] = typeof value === "string" ? value : value();
        });
        const event = {
            id: this.getUUID(),
            date: new Date(),
            result,
            info,
            context: completeContext,
            elapsedTime: 0,
        };
        if (result === "OK") {
            event.action = actionOrErrorCode;
        }
        else {
            event.errorCode = actionOrErrorCode;
        }
        this.logQueue.push(event);
        return () => {
            event.elapsedTime = new Date().getTime() - event.date.getTime();
        };
    }
    getUUID() {
        return (this.uuidCounter++).toString(16);
    }
}
//# sourceMappingURL=EventLogger.js.map