import { put } from "redux-saga/effects";
import { app } from "./app";
import { loadingAction } from "./reducer";
import { stringifyWithMask } from "./util/json";
/**
 * Used for ActionHandler functions
 */
export function createActionHandlerDecorator(interceptor) {
    return (target) => {
        const descriptor = target.descriptor;
        const fn = descriptor.value;
        descriptor.value = function* (...args) {
            const rootState = app.store.getState();
            yield* interceptor(fn.bind(this, ...args), rootState);
        };
        return target;
    };
}
/**
 * Used for regular functions
 */
export function createRegularDecorator(interceptor) {
    return (target) => {
        const descriptor = target.descriptor;
        const fn = descriptor.value;
        descriptor.value = function (...args) {
            const rootState = app.store.getState();
            interceptor(fn.bind(this, ...args), rootState);
        };
        return target;
    };
}
/**
 * Built-in ActionHandler decorators
 */
export function Loading(identifier = "global") {
    return createActionHandlerDecorator(function* (handler) {
        try {
            yield put(loadingAction(true, identifier));
            yield* handler();
        }
        finally {
            yield put(loadingAction(false, identifier));
        }
    });
}
export function Log() {
    return (target) => {
        const descriptor = target.descriptor;
        const fn = descriptor.value;
        descriptor.value = function* (...args) {
            // Do not use fn directly, it is a different object
            const params = stringifyWithMask(app.maskedEventKeywords, "***", ...args);
            const logTypeName = descriptor.value.actionName;
            const context = params ? { params } : {};
            const onLogEnd = app.eventLogger.log(logTypeName, context);
            try {
                yield* fn.bind(this)(...args);
            }
            finally {
                onLogEnd();
            }
        };
        return target;
    };
}
export function Lifecycle() {
    return (target) => {
        const descriptor = target.descriptor;
        descriptor.value.isLifecycle = true;
        return target;
    };
}
export function Interval(second) {
    return (target) => {
        const descriptor = target.descriptor;
        descriptor.value.tickInterval = second;
        return target;
    };
}
/**
 * Built-in regular function decorators
 */
export function Throttle(millisecond) {
    let hasCalled = false;
    return createRegularDecorator(handler => {
        if (!hasCalled) {
            handler();
            hasCalled = true;
            setTimeout(() => {
                hasCalled = false;
            }, millisecond);
        }
    });
}
const defaultMemoKeyGenerator = (args) => JSON.stringify(args);
export function Memo(memoKeyGenerator = defaultMemoKeyGenerator) {
    return (target) => {
        const descriptor = target.descriptor;
        const fn = descriptor.value;
        const cache = {};
        descriptor.value = (...args) => {
            const paramKey = memoKeyGenerator(args);
            if (!cache[paramKey]) {
                cache[paramKey] = fn(...args);
            }
            return cache[paramKey];
        };
        return target;
    };
}
//# sourceMappingURL=decorator.js.map